#CHANGES

#12.1.07: Labels of chains have changed: fist label is kept, last label is a concatenation of label 2..n


#####################################################################################
#   chain.pm                                                              
#                                                                           
#   Copyright (C) 2006 by Magnus Bordewich, Simone Linz, Katherine St. John,
#   and Charles Semple
#
#   This is free software and may be distributed under the same terms as Perl itself.
#        
#   linz@cs.uni-duesseldorf.de
#####################################################################################


package chain;

require tree;


#For n>=3 replace any maximal n-chain that occurs with the same direction in
#both trees T and T' (after applying the chain reduction) by a 2-chain with a labeling representing
#the first and last element of the n-chain. all other taxa can be deleted in the
#WHOLE tree. As we search for maximal n-chains so that no n-chain is part of
#another n-chain the weight of each n-chain is (n-2) for the initial application of the chain reduction
#(not if we apply the chain reduction after the cluster reduction!).

#(Bordewich, Semple 2006)

#Definition of n-chain:
#----------------------
#An n-chain is a set {a_1, a_2, ..., a_n} of leaves such that the parent of each
#a_i is a child of the parent of a_{i+1}, with 1<=i<=(n-1). An n-chain could
#start with a cherry so that the parent of a_1 is the same than the parent of a_2.
#an n-chain can starts with the cherry (a_1, a_2) in either T or T'
#(XOR), but not in both trees because otherwise we have a pendant subtree

#Idea of the algorithm
#---------------------
#(0)Generate an array which includes all nodes of T which are a leaf in T and T' (tree::get_leaves)
#(1)For each pair of leaves (a_i, a_j) with i<j check if both elements presesnt a 
#   2-chain (definition see above) of the same direction in T and T',
#   possible cases:
#   (I)    if the parent of a_i is a child of the parent of a_j in T and T', keep the ordering (a_i, a_j)
#   (II)   if the parent of a_j is a child of the parent of a_i in T and T', swap (dummy=a_i; a_i=a_j; a_j=dummy)
#   (III)  if the parent of a_i is a child of the parent of a_j in T and
#          if the parent of a_i is the same as the parent of a_j in T', keep the ordering (a_i, a_j)
#   (IV)   if the parent of a_j is a child of the parent of a_i in T and
#          if the parent of a_i is the same as the parent of a_j in T', swap (dummy=a_i; a_i=a_j; a_j=dummy)
#   (V)    if if the parent of a_i is the same as the parent of a_j in T and
#          if the parent of a_i is a child of the parent of a_j in T', keep the ordering (a_i, a_j)
#   (VI)   if if the parent of a_i is the same as the parent of a_j in T and
#          if the parent of a_j is a child of the parent of a_i in T', swap (dummy=a_i; a_i=a_j; a_j=dummy)
#   save all 2-chains in a two-dimensional array   
#(2)Repeat
#     join one n-chain (a_1, a_2, ..., a_n) (n>=2) with one m-chain 
#     (b_1, b_2, .., b_m) (m>=2) if one of the following is true:
#         (a) a_n=b_1
#         (b) a_1=b_m
#      The resulting chain has (n+m-1) leaves and is the following
#         for (a) (a_1, a_2, ..., a_{n-1}, b_1, b_2, ...., b_m), delete n-chain and m-chain
#         for (b) (b_1, b_2, ..., b_{m-1}, a_1,a_2, ...., a_n), delete n-chain and m-chain
#   Until no more joining is possible
#(3)For each n-chain (a_1, a_2, ..., a_3) (n>=2, we also keep chains with n=2 and w=0
#   because we can also exclude the edges connecting the parents of the two leaves of
#   these 2-chains from the set of edges which we consider for deletion to find a forest) calculate the weight.
#   If we apply the chain reduction initially in the very beginning the weight of each n-chain is (n-2).
#   In all other cases (after the cluster reduction) (n-2) must be added to the sum over all
#   weights of 2-chains (a,b) whose element are in {a_1, a_2, ..., a_n).
#   (NOTE: each 2-chain can only be in ONE new n-chain!!!)
#(4)Replace each n-chain (a_1, a_2, ..., a_n) (n>2) by a 2-chain (a_1, a_n) and delete all
#   elements in {a_2, a_3, ..., a_{n-1}) in all vertices of T and T'.


#Remark:
#------
#
#(1) organization of set P
#    $set_p[$index_of_two_chain][0]=a_1
#    $set_p[$index_of_two_chain][1]=a_n
#    $set_p[$index_of_two_chain][2]=weight of corresponding n-chain



#finds n-chains in T and T' (both are the inputted trees)
sub find_maximal_chains{
  my @tree1=@{$_[0]};
  my @tree2=@{$_[1]};
  my @old_set_p=@{$_[2]};
  my @weight;
  
  my @set_p=find_two_chains(\@tree1, \@tree2);  
  @set_p=join_chains(\@set_p);  
  @weight=assign_weights(\@set_p, \@old_set_p); 
  @tree1=replace_chains_in_tree(\@set_p, \@tree1);
  @tree2=replace_chains_in_tree(\@set_p, \@tree2);
  @set_p=update_set_p(\@set_p, \@weight);   
  @tree1=tree::sort_tree(\@tree1);
  @tree2=tree::sort_tree(\@tree2); 
  return (\@tree1, \@tree2, \@set_p)
}


#finds 2-chains in T
sub find_two_chains{
  my @tree1=@{$_[0]};
  my @tree2=@{$_[1]};
  my @set_p;
  my $index=0;
  my @leaf=tree::get_leaves(\@tree1);  #T and T' must consist of the same leafset
  for (my $i=0; $i<@leaf; $i++){
      for (my $j=$i+1; $j<@leaf; $j++){
	    my $parent_i=tree::get_parent($leaf[$i], \@tree1);
	    my $parent_j=tree::get_parent($leaf[$j], \@tree1);	 	    
            #print "parent $parent_i $parent_j\n";
	    my @child_i=tree::sons($parent_i, \@tree1);
	    my @child_j=tree::sons($parent_j, \@tree1);
	    #CASES I and III
	    #the parent of i is a child of the parent of j  
            if((($parent_i eq $child_j[0])&&($child_j[1] eq $leaf[$j]))||(($parent_i eq $child_j[1])&&($child_j[0] eq $leaf[$j]))){
	      if(two_chain_in_second_tree($leaf[$i], $leaf[$j], \@tree2)>0){  #doesn't matter if 2-chain in T' is cherry or not
                $set_p[$index][0]=$leaf[$i];
                $set_p[$index][1]=$leaf[$j];
                $index++;	
              }		
            }
	    #CASES II and IV
            #the parent of j is a child of the parent of i    
            elsif((($parent_j eq $child_i[0])&&($child_i[1] eq $leaf[$i]))||(($parent_j eq $child_i[1])&&($child_i[0] eq $leaf[$i]))){
	      if(two_chain_in_second_tree($leaf[$j], $leaf[$i], \@tree2)>0){   #doesn't matter if 2-chain in T' is cherry or not
                $set_p[$index][0]=$leaf[$j];
                $set_p[$index][1]=$leaf[$i];
                $index++;     
              }   
	    }
	    #the parent of i is the same as the parent of j, thus 2-chain is a cherry
	    #CASES V and VI
	    elsif($parent_i eq $parent_j){     	
	      if(two_chain_in_second_tree($leaf[$i], $leaf[$j], \@tree2)==2){  #ordering of first two variables doesn't matter here
	      							   	       #2-chain cannot be a cherry in T'
                #CASE V	
	        #the parent of i is a child of the parent of j in T'
		$set_p[$index][0]=$leaf[$i];
		$set_p[$index][1]=$leaf[$j];						       
		$index++;							  
	      }
	      elsif(two_chain_in_second_tree($leaf[$i], $leaf[$j], \@tree2)==-2){  #ordering of first two variables doesn't matter here
	      								           #2-chain cannot be a cherry in T'
	        ##CASE VI
		#the parent of j is a child of the parent of i in T' 
		$set_p[$index][0]=$leaf[$j];
		$set_p[$index][1]=$leaf[$i];								   
	        $index++;
	      }
	    }
	} 
  }
  return @set_p;
}


#two chain must have same direction as in the first input tree
#two chains in T'
sub two_chain_in_second_tree{
  my $leaf1=$_[0];
  my $leaf2=$_[1];
  my @tree=@{$_[2]};
  my $leaf1_index=tree::find_array_index($leaf1, \@tree);
  my $leaf2_index=tree::find_array_index($leaf2, \@tree);
  my $parent1=tree::get_parent($tree[$leaf1_index], \@tree);
  my $parent2=tree::get_parent($tree[$leaf2_index], \@tree);
  my @child1=tree::sons($parent1, \@tree);
  my @child2=tree::sons($parent2, \@tree);
  if((($parent1 eq $child2[0])&&($child2[1] eq $tree[$leaf2_index]))||(($parent1 eq $child2[1])&&($child2[0] eq $tree[$leaf2_index]))){
    return 2;
  }  
  #only important if a_i and a_j are a cherry in T
  elsif((($parent2 eq $child1[0])&&($child1[1] eq $tree[$leaf1_index]))||(($parent2 eq $child1[1])&&($child1[0] eq $tree[$leaf1_index]))){
    return -2;
  }
  #if 2-chain is a cherry
  elsif($parent1 eq $parent2){
    return 1,
  }
  else{
    return 0;
  }  
  
}

#joins two chains to a new one, if they are neighbours in both trees
sub join_chains{
  my @set_p=@{$_[0]};
  $chain_was_found=1;
  while ($chain_was_found==1){
    $chain_was_found=0;
    my $i=0;
    my $bool=0;
    my $number_of_chains=$#set_p+1;
    while (($bool==0)&&($i<$number_of_chains)){
      my $j=$i+1;
      while (($bool==0)&&($j<$number_of_chains)){
        my $last_chain_index_i=$#{$set_p[$i]};
	my $last_chain_index_j=$#{$set_p[$j]};
	if ($set_p[$i][$last_chain_index_i] eq $set_p[$j][0]){
	  $bool=1;
	  pop(@{$set_p[$i]});
	  push(@{$set_p[$i]}, @{$set_p[$j]});		#concatenate array
	  @{$set_p[$j]}="";
	}
	elsif ($set_p[$i][0] eq $set_p[$j][$last_chain_index_j]){
	  $bool=1;
	  pop(@{$set_p[$j]});
	  push(@{$set_p[$j]}, @{$set_p[$i]});		#concatenate array
	  @{$set_p[$i]}="";
	}
	if ($bool==1){
	  $chain_was_found=1;
	  @set_p=tree::delete_empty_elements_2d_array(\@set_p);
	}
	$j++;
      }
      $i++;
    }
    
  }  
  return @set_p;
}

#calculates the weight of each chain
sub assign_weights{
  my @set_p=@{$_[0]};
  my @old_set_p=@{$_[1]};
  my @weight;
  for my $chain ( 0 .. $#set_p ){ 
    my $taxa_of_maximal_chain=join(",",@{$set_p[$chain]});
    $weight[$chain]=@{$set_p[$chain]}-2;  #thus for 2-chains we have a weight of 0
    #the weight of an n-chain (a_1, a_2, ..., a_3) is not only (n-2), but must be added 
    #to the sum over all weights of 2-chains (a,b) whose element are in {a_1, a_2, ..., a_n)
    #all 2-chains for which this is true are deleted
    #each 2-chain can only be in ONE new n-chain!!!
    if (@old_set_p){  
      for my $old_chain ( 0 .. $#old_set_p ){ 
        #see remark of (Bordewich and Semple 2006): the label set of any maximal n-chain must intersect each pair of P in either both
        #elements or neither
        if (tree::is_in($taxa_of_maximal_chain, $old_set_p[$old_chain][0].",".$old_set_p[$old_chain][1])){
	  $weight[$chain]=$weight[$chain]+$old_set_p[$old_chain][2];
	}
      }
    }
  }  
  return (@weight)
}


#replaces the chain in the tree with two new leaves (first label as label of first chain elem, second label as concatenation of all others)
sub replace_chains_in_tree{
  my @set_p=@{$_[0]};
  my @tree=@{$_[1]};
  for my $chain ( 0 .. $#set_p ){ 
    my @new_label;
    my $i;
    for ($i=1; $i<@{$set_p[$chain]}-1; $i++){   #first and last chain element are not replaced in the tree, thus nothing for a 2-chain
      push(@new_label, $set_p[$chain][$i]);
      #delete the parent node and the leaf itself
      my $parent=tree::get_parent($set_p[$chain][$i], \@tree);
      @tree=tree::delete_node($parent,\@tree);
      @tree=tree::delete_node($set_p[$chain][$i],\@tree);
    }  
    
    my $new_label="";
    $new_label= join(",",@new_label);
    if ($new_label ne ""){
      foreach my $elem (@tree){
        if ($elem eq $set_p[$chain][$i]){
          $elem=$elem.",".$new_label;
        }
      }	
    }
  }    
    
  tree::sort_tree(\@tree);
  return @tree;
}



#replaces any n-chain (a_1, a_2, ..., a_n) with a 2-chain (a_1, a_n) 
#and write weights in the same array
#$set_p[$chain_number][0]=a_1
#$set_p[$chain_number][1]=a_n
#$set_p[$chain_number][2]=weight of corresponding n-chain
sub update_set_p{
  my @set_p=@{$_[0]};
  my @weight=@{$_[1]};
  my @new_set_p;
  for (my $i=0; $i<=$#set_p; $i++){
    $new_set_p[$i][0]=$set_p[$i][0];
    my @label_set;
    for (my $j=1; $j<=$#{$set_p[$i]}; $j++){
      push(@label_set, $set_p[$i][$j]);      
    }
    my $label=join(",",@label_set);
    $label=tree::sort_node($label);
    #$new_set_p[$i][1]=$set_p[$i][$#{$set_p[$i]}];  #!
    $new_set_p[$i][1]=$label;
    $new_set_p[$i][2]=$weight[$i];
  }
  tree::print_2d_array("set_p", \@new_set_p);
  return @new_set_p;
}
  

1;



