#####################################################################################
#   cluster.pm                                                              
#                                                                           
#   Copyright (C) 2006 by Magnus Bordewich, Simone Linz, Katherine St. John,
#   and Charles Semple
#
#   This is free software and may be distributed under the same terms as Perl itself.
#        
#   linz@cs.uni-duesseldorf.de
####################################################################################

#divide and conquer approach to find the minimum hybridization number for two given rooted binary phylogenetic trees  
#by breaking up the initial two trees into a set of minimal clusters (same leaf set, but different topology). 
#the final number of hybridization events is the sum over all numbers of hybridization for each cluster pair.

#(Baroni, Semple, Steel 2006)



package cluster;


#finds a minimal cluster, in a way that the cluster consists of the same taxa
#in T and T' but the topology of both subtrees is different
sub find_cluster{
	my @tree1=@{$_[0]};
	my @tree2=@{$_[1]};
	%tree2=tree::array2hash(\@tree2);
	my $cluster="";
	my $i=@tree1-1;
	while (($cluster eq "")&&($i>=0)){ 
		if ((not tree::is_leaf($tree1[$i], \@tree1))&&(exists $tree2{$tree1[$i]})){   #because T and T' consist of the same set of taxa											
			$cluster=$tree1[$i];
		}
		$i--;
	}

	my @cluster_tree1=get_cluster_tree($cluster, \@tree1);
	my @cluster_tree2=get_cluster_tree($cluster, \@tree2);
	
	@tree1=update_tree($cluster, \@tree1);
	@tree2=update_tree($cluster, \@tree2);
	#my $no=tree::count_leaves(\@cluster_tree1);
	#print "no of leaves in cluster : $no\n\n";
	return (\@cluster_tree1, \@cluster_tree2, \@tree1, \@tree2);
}


#extract the minimal cluster from a tree
sub get_cluster_tree{
	my $cluster=$_[0];
	my @tree=@{$_[1]};
	my @cluster_tree;
	my $index=0;
	foreach my $node (@tree){
		if(tree::is_in($cluster, $node)){
			$cluster_tree[$index]=$node;
			$index++;
		}
		
	}
	return @cluster_tree;
}


#cuts the cluster from the input tree T: T \ taxa of cluster, replacing the cluster with a single vertex 
#representing all taxa of the cluster
sub update_tree{
	my $cluster=$_[0];
	my @tree=@{$_[1]};
	my $bool=0;

	foreach my $node (@tree){
		if ($bool==0){
			if ($node eq $cluster){
				$bool=1;
			}
		}
		else{
			if(tree::is_in($cluster, $node)){
				$node="";
			}
		}
			
	}
	@tree=tree::delete_empty_elements(\@tree);
	@tree=tree::delete_double_elements(\@tree);
	@tree=tree::sort_tree(\@tree);
	return @tree;
}


#every n-chain, reduced to a 2-chain still stays together after the cluster 
#reduction. e.g. if T and T' have one common 3-chain {a,b,c} which is replaced by {a,c} 
#we have P={{a,c}} with w({a,c})=1. splitting up T and T' (after applying the subtree and chain reduction rule) 
#into MINIMAL clusters one cluster consists of at least BOTH taxa 'a' and 'c'. 
#it is NOT possible that one cluster consists of at least 'a', but not 'c' while another 
#cluster consits of at least 'c', but not 'a'.
sub update_set_p{
  my @set_p=@{$_[0]};
  my @cluster=@{$_[1]};
  my @current_set_p;
  my $index=0;
  for my $chain ( 0 .. $#set_p ){ 
    if((@{$set_p[$chain]} ne "")&&(tree::is_in($cluster[0], $set_p[$chain][0]))&&(tree::is_in($cluster[0], $set_p[$chain][1]))){
       @{$current_set_p[$index]}=@{$set_p[$chain]};
       $index++;
       @{$set_p[$chain]}="";
       
    }
    elsif (((tree::is_in($cluster[0], $set_p[$chain][0]))&&(not tree::is_in($cluster[0], $set_p[$chain][1])))||((not tree::is_in($cluster[0], $set_p[$chain][0]))&&(tree::is_in($cluster[0], $set_p[$chain][1])))){
      print "ERROR cluster::update_set_p\n\n";
      die;
    }
  }
  @set_p=tree::delete_empty_elements_2d_array(\@set_p);
  tree::print_2d_array("test", \@current_set_p);
  return (\@set_p, \@current_set_p)
}



1;
