####################################################################################
#   forest.pm                                                              
#                                                                           
#   Copyright (C) 2006 by Magnus Bordewich, Simone Linz, Katherine St. John,
#   and Charles Semple
#
#   This is free software and may be distributed under the same terms as Perl itself.
#        
#   linz@cs.uni-duesseldorf.de
####################################################################################

package forest;

use List::Compare;
use Graph::Directed;
use POSIX;
use Array::Compare;

#Short description of the idea to find an appropriate forest:
#############################################################
#(1)delete n edges of tree1 and calculate the resulting (n+1) subtrees
#(2)add root to tree2
#(3)get leafsets for all (n+1) subtrees (note: leaf 3,4 will be represented with 3 and 4 in leafset)
#(4)for each leafset get minimal subtree in tree2, without supressing degree-2 nodes
#eg: #1,2,3,4 1,2,3,4 1,2,3 1,2 1 2 3 4 minimal subtree for 1 and 3 is 1,2,3 1,2 1 3
#(5)for all minimal subtrees do a pairwise check, if all intersections are empty
#(6)check if the real subtrees (restricted subtrees of tree2) are the same and check if the
#forest is acyclic. if so: forest found!!

#extension for finding a maximum-legitimate-acyclic-agreement forest (Bordewich, Semple 2006), 
#considering of the weights of n-chains, and a current best hybridization number
#implementation of exhaustive search 
#      (1) don't delete all three edges incident with one vertex
#      (2) don't delete edges seperating the parents of the leaves in a 2-chain 
#      (3) don't cut edges in a way, such that the root rho (#) is isolated in the forest 
#	     (Baroni, Gruenewald, Moulton, Semple 05, Lemma 1)



sub find_forest{
  my @tree_array1=@{$_[0]};
  my @tree_array2=@{$_[1]};
  my @set_p=@{$_[2]};
  
  my $current_best=INT_MAX;
  my $extra_score;
  my @tree_to_cut;

  #agreement forest is equal to the input tree if both trees are the same
  if (tree::compare_arrays(\@tree_array1, \@tree_array2)){
    return 0;
  }
	
  else{
    my $no_leaves=tree::count_leaves(\@tree_array1);
    
    #generate array with all leaves
    #for checking if root node is not isolated in the forest
    my @leaf_set_tree=split(/\,/, $tree_array1[0]);
    
    #intersection of L(T_p) and the label set X is not empty (Baroni, Gruenewald, Moulton, Semple 05, Lemma 1)
    @tree_to_cut=tree::delete_node($tree_array1[0], \@tree_array1);
    
    #delete edges seperating the parents of the leaves in each 2-chain with weight>0
    @tree_to_cut=delete_parent_sep(\@tree_array1, \@tree_to_cut, \@set_p);
    
    my @original_tree=reverse(@tree_to_cut);
    @tree_to_cut=@original_tree;
    
    my $global_bool; #checks if there are any more combinations with i cuts in the tree
    my $root_isolated=0;
    my $forest_found=0;

    #delete edges in tree and test if subtrees are an agreement forest, the resulting forest is equal to the max agreement
    #forest because we increase the number of cuts in a tree starting with 1 cuts (i cuts => delete i edges)
    for (my $cut=1; $cut<=$no_leaves; $cut++){  #worst case: all leaves and the root are isolated vertices in the forest
      print "Checking $cut edge(s) to delete!\n";
      if ($current_best==$cut){
        return $cut;
      }
      else{	 
        @tree_to_cut=@original_tree;
        my @index;
        my @bool;  #bool value for each position
	
	for (my $j=0; $j<$cut; $j++){  #initialize: first combination for each number of cuts
          $index[$j]=$j;
          @bool=adapt_bool_value($j, $cut, \@bool, \@tree_to_cut, \@index);
          @tree_to_cut=modify_tree($index[$j], \@tree_to_cut);
        }
        $root_isolated=intersect_leaf_sets(\@leaf_set_tree, \@index, \@tree_to_cut);
	if ($root_isolated==0){
	  $extra_score=check_combination($cut, \@tree_to_cut, \@index, \@set_p, \@tree_array1, \@tree_array2);
	  if ($extra_score==0){
	    return $cut;
	  }
	  elsif (($extra_score>0)&&($current_best>$extra_score+$cut)){
            $current_best=$cut+$extra_score;
	    $forest_found=0;
	    #print "current best update $current_best\n";
	  }
	  #else extra_score=-1  ->no forest with this combinations of edges to delete 
        }
		
	$global_bool=test_for_more_combinations(\@bool); #if $global_bool==1 --> more combinations exist
        while($global_bool==1){
	  @tree_to_cut=@original_tree;
          my $i=@bool;
          my $found=0; 
          while($found==0){ 
            $i--;
            if ($bool[$i]==1){
              $found++;
            }
          }
    
          for ($j=0; $j<$i; $j++){
            @tree_to_cut=modify_tree($index[$j], \@tree_to_cut);
          }
    
          $index[$i]++; 
    
          @bool=adapt_bool_value($i, $cut, \@bool, \@tree_to_cut, \@index);
          @tree_to_cut=modify_tree($index[$i], \@tree_to_cut);
      
          #adjust indices 
          my $k=$index[$i]; 
          my $def=1;
          $j=$i+1;
	  
          while(($def==1)&&($j<$cut)){
            $k++;
            if (defined $tree_to_cut[$k]){
              $index[$j]=$k;  
              @bool=adapt_bool_value($j, $cut, \@bool, \@tree_to_cut, \@index); 
              @tree_to_cut=modify_tree($index[$j], \@tree_to_cut);
	      $def=1;
            }
            else{
              $def=0;
            }
            $j++;
          }
    
          if ($def==1){
            $root_isolated=intersect_leaf_sets(\@leaf_set_tree, \@index, \@tree_to_cut);
            if ($root_isolated==0){
	      $extra_score=check_combination($cut, \@tree_to_cut, \@index, \@set_p, \@tree_array1, \@tree_array2);
	      if ($extra_score==0){
	        return $cut;
	      }
	      elsif (($extra_score>0)&&($current_best>$extra_score+$cut)){
                $current_best=$cut+$extra_score;
	        $forest_found=0;
	        #print "current best update $current_best\n";
	      }
	     #else extra_score=-1  ->no forest with this combinations of edges to delete 
            }
          }  
          $global_bool=test_for_more_combinations(\@bool);
        }
		
      }	
    }
  }  
}   



#checks combinations of edges to delete
#if forest exists, this function returns $extra_score, otherwise -1
sub check_combination{
  my $no_cuts=$_[0];
  my @tree_to_cut=@{$_[1]};;
  my @index=@{$_[2]};
  my @set_p=@{$_[3]};
  my @tree_array1=@{$_[4]};
  my @tree_array2=@{$_[5]};
  my @cut; #array with edges to delete
  
  my $bool;
  
  #print "$cut cuts:";
  
  foreach my $elem (@index){
    push (@cut, $tree_to_cut[$elem]);
    #output#####################
    #print " $tree_to_cut[$elem]";
  }
  #print "\n"; 
  
  #extra_score=-1 --> if there is only one leaf of any 2-chain isolated; there is no need to check whether an 
  #appropriate forest exists because the two leaves of a chain either stay together in one component of the forest or each
  #leaf is isolated in the forest
  #if both leaves are snipped the weight of the chain must be added to the forest size minus 1 (see function: calculate_score)
  $extra_score=calculate_score(\@set_p, \@cut);  
  if ($extra_score>-1){
    $bool=forest(\@cut, \@tree_array1, \@tree_array2);
    if ($bool==1){
      return $extra_score;
    }
    else{
      return -1;
    }  
  }  
  else{
    return -1;
  }  
}


#checks if an appropriate forest exists for a valid combination of edges to delete
sub forest{
      my @tree1_cut=@{$_[0]};
      my @tree_array1=@{$_[1]};
      my @tree_array2=@{$_[2]};
      my @forest1;
      my @forest2;
      my @subtree1;
      my @subtree2;
      my $local_bool;
      $bool=0;

      #write subtrees after deleting edges in array 
      @subtree1=write_subtrees(\@tree_array1, \@tree1_cut);
      
      #calculate leafsets for all subtrees
      my @leafset; #array representing the leaf set of each tree in the forest
      for (my $curr_subtree_no=0; $curr_subtree_no<@subtree1; $curr_subtree_no++){
	#leafclusters are also splitted
	$leafset[$curr_subtree_no]=$subtree1[$curr_subtree_no][0];
      } 
      
      #find minimal subtrees in tree2, without supressing degree 2 vertices  
      @tree_array2=tree::insert_root(\@tree_array2);
      for (my $curr_leafset_no=0; $curr_leafset_no<@leafset; $curr_leafset_no++){
	@{$subtree2[$curr_leafset_no]}=get_minimal_subtree($leafset[$curr_leafset_no], \@tree_array2);
      }	
      
      if (@subtree1==@subtree2){  #not necessary because we don't delete all three edges incident with one vertex
        #pairwise check if subtree2 trees are vertex disjoint?
        $local_bool=1;
	my $k=0;
	while (($local_bool==1)&&($k<@subtree2)){  
		my $l=$k+1;
		while(($local_bool==1)&&($l<@subtree2)){ 
			my $lc = List::Compare->new('-u', \@{$subtree2[$k]}, \@{$subtree2[$l]});
			my @intersection = $lc->get_intersection;
			if (defined($intersection[0])){
				$local_bool=0;
			}
			$l++;
	        }
		$k++;
        }	
	
        if ($local_bool==1){
	  #delete degree two vertices in subtree2 and taxa which are not belonging to the leafset
	  for (my $k=0; $k<@subtree2; $k++) { 
	    @{$subtree2[$k]}=get_restricted_subtree($leafset[$k], \@{$subtree2[$k]});
	  }	
	      
	  #check if both subtree arrays are the same
	  @forest1=forest_as_string(\@subtree1);
	  @forest2=forest_as_string(\@subtree2);
	  if(forests_are_equal(\@forest1, \@forest2)){
	  
	    #check if graph is acyclic, if so: forest found
	    my $graph=make_graph(\@subtree1, \@tree_array1, \@tree_array2);
	    if ($graph eq "0"){  #cycles in graph
	      #print "graph is cyclic\n\n";
	    }
 	    else{
	      #print "graph is not cyclic\n\n";
	      #print "forest found\n";
	      $bool=1;
	    }  
 	  }
        }
      }  
      return $bool;     
}


#sub sort_cuts{
#  my @tree=@{$_[0]};
#  my @cut_position=@{$_[1]};
#  my @cut_position_sorted;
#  return @cut_position_sorted;
#}
#@tree1_cut_sorted=sort_cuts(\@tree_array1, \@tree1_cut);



#deletes elements of the tree array which will never get cutted
#in this case:
#there is no need to delete edges seperating the parents of the leaves in a 2-chain {a,b} with w({a,b})>0 and in 
#which (a,b) is NOT a cherry because to obtain a forest either a and b are both contained in the label set of ONE 
#tree in the forest or a and b label isolated vertices in the forest
sub delete_parent_sep{
  my @tree=@{$_[0]};
  my @tree_to_cut=@{$_[1]};
  my @set_p=@{$_[2]};
  
  for my $chain ( 0 .. $#set_p ){ 
    if ($set_p[$chain][2]>0){
      my $parent_one=tree::get_parent($set_p[$chain][0], \@tree);
      my $parent_two=tree::get_parent($set_p[$chain][1], \@tree);
      if ($parent_two eq tree::get_parent($parent_one, \@tree)){
        @tree_to_cut=tree::delete_node($parent_one, \@tree_to_cut);
      }
      elsif($parent_one eq tree::get_parent($parent_two, \@tree)){
        @tree_to_cut=tree::delete_node($parent_two, \@tree_to_cut);
      }
      elsif($parent_one eq $parent_two){
        #2-chain is a cherry, nothing to do!!
      }
      else{
        print "forest::delete_parent_sep parents ERROR: something wrong with parents of leaves in 2-chain\n";
        die;
      }
    }
  }
  return @tree_to_cut;
}



#checks if leaves of a reduced 2-chain stay together in one component of the forest or are isolated vertices. 
#otherwise this combination of edges to delete needn't to be considered any further. if both leaves are isolated vertices
#in the forest the weight of the 2-chain must be added to the number of trees in the forest minus 1
sub calculate_score{
  my @set_p=@{$_[0]};
  my @cut_pos=@{$_[1]};
  my %cut_pos=tree::array2hash(\@cut_pos);
  my $score=0;
  my $bool=1;
  for my $chain ( 0 .. $#set_p ){ 
    if ($set_p[$chain][2]>0){   #necessary because of (Bordewich, Semple 2006, Lemma 3.1.)
    				#otherwise: T 1,2,3 1,2 1 2 3 and T' 1,2,3 1,3 1 2 3 with h(T,T')=2 (instead of 1)!!
      if ((exists $cut_pos{$set_p[$chain][0]})&&(exists $cut_pos{$set_p[$chain][1]})){
        $score=$score+$set_p[$chain][2];  
      }  
      elsif(((!exists $cut_pos{$set_p[$chain][0]})&&(exists $cut_pos{$set_p[$chain][1]}))||((exists $cut_pos{$set_p[$chain][0]})&&(!exists $cut_pos{$set_p[$chain][1]}))){
        $bool=-1;
      }	
    }
  }
  if ($bool==1){
    return $score;
  }
  elsif ($bool==-1){
    return $bool;
  }  
}



#gives back a minimal subtree for a set of leaves
#WITHOUT supressing degree two vertices
sub get_minimal_subtree{
	my $leafset=$_[0];
	my @tree=@{$_[1]};
	my @subtree;
	my $bool=0;
	my $i=@tree-1;
	my @leafset_array=split(/\,/, $leafset);
	
	while(($bool==0)&&($i>=0)){  
		my @nodes=split(/\,/, $tree[$i]);
		my $lc = List::Compare->new('-u', \@nodes, \@leafset_array);
		my @intersection = $lc->get_intersection;
		if (defined($intersection[0])){
			push (@subtree, $tree[$i]);
			#check if all leaves are in the current node
			$bool=smallest_subtree_root($tree[$i], \@leafset_array);
			
		}
		$i--;	
	}
	@subtree=reverse(@subtree);
	return @subtree;

}



#checks if a vertex comprises at least all taxa of a given set of leaves
sub smallest_subtree_root{
	my $node=$_[0];
	my @leafset=@{$_[1]};
	my $bool=1;
        my @node=split(/\,/, $node);
        my %node=tree::array2hash(\@node);
	my $index=0;
	while (($bool==1)&&($index<@leafset)){
		if (! exists $node{$leafset[$index]}){
			$bool=0
		}
		$index++;
	}
	return $bool;
}



#gives back a minimal subtree for a set of leaves
#WITH supressing degree two vertices
sub get_restricted_subtree{
	my $leafset=$_[0];
	my @tree=@{$_[1]};
	my @leafset_array=split(/\,/, $leafset);
	my %leafset_hash=tree::array2hash(\@leafset_array);
	foreach my $elem (@tree){
		my @taxa=split(/\,/, $elem);
		foreach my $elem2 (@taxa){
			if(! exists $leafset_hash{$elem2}){
				$elem2="";
			}
		}
		$elem=join(",", @taxa);
		$elem=tree::remove_commas($elem);
	}
	@tree=tree::delete_empty_elements(\@tree);
	@tree=tree::delete_double_elements(\@tree);
	@tree=tree::sort_tree(\@tree);
	return @tree;
}


#transforms a 2d array, containing a forest, into a 1d array, containing each tree as a string (a space seperates the nodes)
sub forest_as_string{
	my @forest_in=@{$_[0]};
	my @forest_out;
	for my $elem ( 0 .. $#forest_in ) { 
		my $string=join(" ", @{$forest_in[$elem]});
		push(@forest_out, $string);
	}
	return @forest_out;
}



#deletes on edge in the tree-array and gives back the subtree which has evolved lastly
#e.g.: abcdefgh abcd efgh ab cd ef gh a b c d e f g h 
#      cut at array position 6 (gh) --> gh g h is the corresponding subtree
#(opposite to function tree_without_subtree)
sub cut_tree{
	my $cut_position=$_[0];  #knoten index VOR dem der schnitt stattfindet
	my @tree=@{$_[1]};
        my @subtree;
	for (my $i=$cut_position; $i<@tree; $i++){
		if(tree::is_in($tree[$cut_position], $tree[$i])){
		    push(@subtree, $tree[$i]);
		}
	}
	@subtree=tree::sort_tree(\@subtree);
	return @subtree;
}



#gives back the remaining tree, without subtree (opposite to function cut_tree)
sub tree_without_subtree{
	my $node=$_[0];
	my @tree=@{$_[1]};
        my @taxa=split(/\,/, $node);
        foreach my $elem (@tree){
	    foreach my $elem2 (@taxa){
		    $elem=tree::substitute_node($elem, $elem2, "");
	    }
        }
	
	
	@tree=tree::delete_empty_elements(\@tree);
	@tree=tree::delete_double_elements(\@tree);

	@tree=tree::sort_tree(\@tree);
	return @tree;
}



#writes all tree in forest in a 2D array
sub write_subtrees{
	my @tree=@{$_[0]};
	my @tree_cut_position=@{$_[1]};
        my @subtree;
        my $subtree_index=-1;

       for (my $i=0; $i<@tree_cut_position; $i++){
		if (defined($tree_cut_position[$i])){
			my $index=tree::find_array_index($tree_cut_position[$i], \@tree);
			$subtree_index++,
			@{$subtree[$subtree_index]}=cut_tree($index, \@tree);
			@tree=tree_without_subtree($tree_cut_position[$i], \@tree);	
		}
	}
	#add the last remaining tree to subtree
	$subtree_index++;  
	@tree=tree::insert_root(\@tree);
	@{$subtree[$subtree_index]}=@tree;
	
	return (@subtree);
}


#checks if the two obtained forests (one of each tree) are equal (one array of trees must be a permutation of the second one)
sub forests_are_equal{
	my @subtree1=@{$_[0]};
	my @subtree2=@{$_[1]};
	my $comp = Array::Compare->new;  
       	if ($comp->perm(\@subtree1, \@subtree2)) {
	       	return 1;
       	}
       	else{
		return 0;
       	}   
}



#transforms the maximum agreement forest into an directed graph, with nodes corresponding to the trees in the forest
#as we are only interested in the root of each tree we only store the root node in the graph instead of the 
#whole tree for each node
#check if the resulting graph is acyclic (see Baroni, Gruenewald, Moulton, Semple 2005)
sub make_graph{
	my @forest=@{$_[0]};
	my @tree1=@{$_[1]};
	my @tree2=@{$_[2]};
	
	my $graph=Graph::Directed->new;
	
	#add vertices to graph
	for my $tree_index ( 0 .. $#forest ) { 
		$graph->add_vertex($forest[$tree_index][0]);
	}
	#add edges to graph
	for my $i ( 0 .. $#forest ) { 
		for my $j ( 0 .. $#forest ) { 
			if ($i!=$j){
				my $bool=is_ancestor($forest[$i][0], $forest[$j][0], \@tree1, \@forest);
				if ($bool==1){
					#print "maximum good/acylick agreement forest edge: $forest[$i][0] $forest[$j][0]\n";
					$graph->add_edge($forest[$i][0], $forest[$j][0]);
				}
				$bool=is_ancestor($forest[$i][0], $forest[$j][0], \@tree2, \@forest);
				if ($bool==1){
					#print "maximum good/acyclic agreement forest edge: $forest[$i][0] $forest[$j][0]\n";
					$graph->add_edge($forest[$i][0], $forest[$j][0]);
				}
			}
		}
		
	}
	#print "\n";
	#find a cycle
	if($graph->has_a_cycle){
		#print "CYCLIC GRAPH $graph\n\n";
		return 0;
	}
	else{
		#print "no cycle\n";
		return $graph;
	}
}



#checks if the root of the tree with index i in the forest  is an ancestor of the root of 
#the tree with index j in the forest
sub is_ancestor{
	my $node_i=$_[0];
	my $node_j=$_[1];
	my @tree=@{$_[2]};
	my @forest=@{$_[3]};
	
	@tree=tree::insert_root(\@tree);
	my $tree_node_i=get_node_in_tree($node_i, \@tree);
	my $tree_node_j=get_node_in_tree($node_j, \@tree);
	if (tree::is_in($tree_node_i, $tree_node_j)){
		return 1;
	}
	else{
		return 0;
	}

}



sub get_node_in_tree{
	my $taxa=$_[0];
	my @tree=@{$_[1]};
	my $node="";
	if ($taxa=~/^\#/){
		$node=$tree[0];
	}
	else{
		my $bool=0;
		for (my $i=@tree-1; $i>0; $i--){
			if (($bool==0)&&(tree::is_in($tree[$i], $taxa))){
				$bool=1;
				$node=$tree[$i];
			}
		}
	}
	if ($node eq ""){
		print "ERROR max_agreement_forest::get_node_in_tree could not find a node\n\n";
		die;
	}
	else{
		return $node;
	}
}




#####################subs for exhaustive search####################################

#checks if there are any further combinations need to be tested? (if so, return 1, else 0)
sub test_for_more_combinations{
  my @bool=@{$_[0]};
  my $global_bool=0;
  foreach my $elem (@bool){
    if ($elem==1){
      $global_bool=1;
    }
  }
  return $global_bool;
}



sub adapt_bool_value{
  my $i=$_[0];
  my $cut=$_[1];
  my @bool=@{$_[2]};
  my @tree=@{$_[3]};
  my @index=@{$_[4]};
  if ($index[$i]<@tree-$cut+$i){
    $bool[$i]=1;  #zeiger kann noch weiter nach rechts verschoben werden
  }
  else{
    $bool[$i]=0;  #zeiger kann nicht noch weiter nach rechts verschoben werden
  }
  return @bool;
}



sub modify_tree{
  my $index=$_[0];
  my @tree=@{$_[1]};
  for (my $i=$index+1; $i<@tree; $i++){
    $tree[$i]=tree::delete_part_of_string($tree[$i], $tree[$index]);
  }
  @tree=tree::delete_empty_elements(\@tree);
  @tree=tree::delete_double_elements(\@tree);
  return @tree
}



#checks if the root is isolated in the forest (Baroni, Gruenewald, Moulton, Semple 2005 --> Lemma 1)
#if the root is isolated this function returns 1, otherwise 0
sub intersect_leaf_sets{
  my @leaf_set_tree=@{$_[0]};
  my @index_edges_to_delete=@{$_[1]};
  my @tree=@{$_[2]};
  my $leaf_set_forest_string="";
  foreach my $elem (@index_edges_to_delete){
    $leaf_set_forest_string=$leaf_set_forest_string.",".$tree[$elem];
  }
  $leaf_set_forest_string=tree::remove_commas($leaf_set_forest_string);
  my @leaf_set_forest=split(/\,/, $leaf_set_forest_string);
  if (@leaf_set_forest==@leaf_set_tree){
    return 1;
  }
  else{
    return 0;
  }
}

1;
