#####################################################################################
#   subtree.pm                                                              
#                                                                           
#   Copyright (C) 2006 by Magnus Bordewich, Simone Linz, Katherine St. John,
#   and Charles Semple
#
#   This is free software and may be distributed under the same terms as Perl itself.
#        
#   linz@cs.uni-duesseldorf.de
#####################################################################################

package subtree;

use tree;

#replace any maximal pendant subtree that occurs identically in T and T' by a single leaf with a new label
#(label consists of the root of the pendant subtree)
#if set P is not empty (this is only the case when we apply the subtree reduction after the cluster reduction) delete all
#members of P whose elements label leaves of the pendant subtree. because of the remark (Bordewich, Semple 2006, page7) it is 
#sufficient to test for each 2-chain (a,b) if the maximal pendant subtree consists of at least 'a'

#(Bordewich, Semple 2006)

#input: two tree arrays (T, T')
#output: T and T' in which subtrees that occur in both input trees are 
#replaced by their most recent common ancestor
#e.g. input tree1 abcdefgh abcd efgh ab cd ef gh a b c d e f g h
#     input tree2 abcdefgh abcd efgh ab cd eg fh a b c d e f g h
#     output      abcdefgh abcd efgh ef gh e f g h

#Idea of algorithm:
#------------------
#(1)for each node of T which is not empty ("") and is also a node in T'
#calculate an array which contains of all descendants in T. check if each array element exists in T' too. if yes,
#repalce each node in T and T' which is equal to an array element with "".
#(2)delete all "" nodes in T and T'
#(3)sort T and T'
#(4)if P is not empty (see above) delete all members of P whose elements label leaves of the maximal pendant subtree


sub find_pendant_subtrees{
  my @tree1=@{$_[0]};
  my @tree2=@{$_[1]};
  my @set_p=@{$_[2]};
  
  my %hash2=tree::array2hash(\@tree2);
  for (my $i=1; $i<@tree1; $i++){  #root must be equal
	if($tree1[$i] ne ""){ 	
		if (exists $hash2{$tree1[$i]}){
			my @subtree_array=get_restriction($i, \@tree1);
			if (@subtree_array>1){ #otherwise only a leaf which is not replaced!!!!!!
	    		  #test if all subtree elements exists in the second tree
	  		  if (tree::is_array_in_hash(\@subtree_array, \%hash2)){
		  		my %subtree_hash=tree::array2hash(\@subtree_array);
		  		#reduce pendant subtree by a single leaf (root of the pendant subtree)
				@tree1=label_new(\@tree1, \%subtree_hash);
				@tree2=label_new(\@tree2, \%subtree_hash);
				if (@set_p){
				  @set_p=update_set_p($tree1[$i], \@set_p);
				} 
			  }	 
	  		}
  		}  
             			
  	  }
  }
  @tree1=tree::delete_empty_elements(\@tree1);
  @tree1=tree::sort_tree(\@tree1);
  @tree2=tree::delete_empty_elements(\@tree2);
  @tree2=tree::sort_tree(\@tree2);
  return (\@tree1, \@tree2, \@set_p);
    
}


#generates the tree restriction of all taxa of array[index] WITHOUT array[index] itself (!)
#e.g: input array abcd ab cd a b c d
#     index 2
#     output array c d
sub get_restriction{
	my $index=$_[0];
	my @tree=@{$_[1]};
	my @subtree_array;
	my $subtree_array_index=0;
   	my $root_of_subtree=$tree[$index];
	
	for (my $j=$index+1; $j<@tree; $j++){
		if ($tree[$j] ne ""){
			if(tree::is_in($tree[$index], $tree[$j])){
				$subtree_array[$subtree_array_index]=$tree[$j];
			 	$subtree_array_index++;
			}
		}
	}
	return @subtree_array;
}


#replaces pendant subtree with a single leaf which is the root of that subtree
#empty elements of the tree array are set to an empty string "", instead of deleting the whole array element
#because than we have to iterate over the whole array every time what is more time consuming
sub label_new{
	my @tree=@{$_[0]};
	my %subtree=%{$_[1]};
	foreach my $elem (@tree){
		if (($elem ne "")&&(exists $subtree{$elem})){
			$elem="";
		}
	}
	return @tree;
}


#deletes all members of P whose elements label leaves of the maximal pendant subtree
sub update_set_p{
  my $root_of_subtree=$_[0];
  my @set_p=@{$_[1]};
  #the label set of any maximal pendant subtree must intersect each pair of P in either both
  #elements or neither
  for my $chain ( 0 .. $#set_p ){ 
    if((@{$set_p[$chain]} ne "")&&(tree::is_in($root_of_subtree, $set_p[$chain][0]))){  
      @{$set_p[$chain]}="";
    }
  }
  @set_p=tree::delete_empty_elements_2d_array(\@set_p);
  return @set_p;
}

1;
