######################################################################################
#   tree.pm                                                              
#                                                                           
#   Copyright (C) 2006 by Magnus Bordewich, Simone Linz, Katherine St. John,
#   and Charles Semple
#
#   This is free software and may be distributed under the same terms as Perl itself.
#        
#   linz@cs.uni-duesseldorf.de
######################################################################################

package tree;

use Graph::Undirected;



#calculates the two/one/zero sons of each elem
#INPUT: SORTED array consisting of a tree and a node
#OUTPUT: array which contains the descendants (here: sons) of the input array element
sub sons{
    my $node=$_[0];
    my @tree=@{$_[1]};
    
    my @node_taxa=split(/\,/, $node);
    my @sons;
    my $i=0;
    foreach my $elem (@tree){
	     my @elem_taxa=split(/\,/, $elem);
		 if (($i==0)&&(@node_taxa>@elem_taxa)&&(is_in($node, $elem))){
		    $sons[0]=$elem;
		    $i++;  
		    my @taxa=split(/\,/, $elem);
		    foreach my $label (@taxa){
			    $node=substitute_node($node, $label, "");
		    } 
	    }
	    elsif (($i==1)&&($node eq $elem)){
		    $sons[1]=$elem;
	    }
    }
	return @sons;
}



#substitutes in the first string the second string with the third string
#input: 3 strings
#output:  1 string
sub substitute_node{
	my $string=$_[0];
	my $to_substitute=$_[1];
	my $substitute_with=$_[2];
	
	my @taxa=split(/\,/, $string);
	foreach my $elem (@taxa){
		if ($elem eq $to_substitute){
			$elem=$substitute_with;
		}
	}
	$string=join(",", @taxa);
	$string=remove_commas($string);
	return $string;
}



#substitutes in in every element of the input tree (array) the first string with the second string
#input: 2 strings and the tree-array
#output: tree-array
sub substitute_tree{
	my $to_substitute=$_[0];
	my $substitute_with=$_[1];
	my @tree=@{$_[2]};
	foreach my $node (@tree){
		$node=substitute_node($node, $to_substitute, $substitute_with);
	}
	@tree=delete_empty_elements(\@tree);  #in case to substitute with nothing
	@tree=delete_double_elements(\@tree);
	@tree=sort_tree(\@tree);
	return @tree;
}



#deletes every empty element in an array
#INPUT: array
#OUTPUT: array
sub delete_empty_elements{
	my @array=@{$_[0]};
	my @new_array;
	my $new_array_index=-1;
	foreach my $elem (@array){
		if ($elem ne "") {
			$new_array_index++;
			$new_array[$new_array_index]=$elem;
		}
	}
	@tree=sort_tree(\@tree);
	return @new_array;
}



#deletes every empty element in an 2d array (empty if array[i][0]=""
#INPUT: 2d array
#OUTPUT: 2d array
sub delete_empty_elements_2d_array{
	my @array=@{$_[0]};
	my @new_array;
	my $new_array_index=-1;
	for my $elem ( 0 .. $#array ) { 
	     if ($array[$elem][0] ne ""){
	         $new_array_index++;
		 @{$new_array[$new_array_index]}=@{$array[$elem]};
	     }
	}
	return @new_array;
}



#makes all array elements unique, by deleting doubles
#eg: array 1 2 4 0 2 7 7
#    output 1 2 4 0 7
#INPUT: array
#OUTPUT: array
sub delete_double_elements{
	my @array=@{$_[0]};
	my @new_array;
	my $new_array_index=0;
	for (my $i=0; $i<@array; $i++){
		my $bool=1;
		for (my $j=0; $j<@new_array; $j++){
	    	if ($array[$i] eq $new_array[$j]){
		    	$bool=0;
	    	}
		}
		if ($bool==1){
			$new_array[$new_array_index]=$array[$i];
			$new_array_index++;
		}
	}
	return @new_array;
}
	


#finds the index of an arrayelement in an array
#INPUT: array and one array element
#OUTPUT: index of the array element in array
sub find_array_index{
	my $array_elem=$_[0];
	my @array=@{$_[1]};
	my $index=-1;
	for (my $i=0; $i<@array; $i++){
		if ($array[$i] eq $array_elem){
			$index=$i;
		}
	}
	if ($index==-1){
		return  0;
	}
	return $index;
}



#compares two arrays (one dimension) and gives back 1 if they are equal
#INPUT: two SORTED arrays
#OUTPUT: boolean
sub compare_arrays { 
    my ($first, $second) = @_; 
    no warnings;  # silence spurious -w undef complaints 
    return 0 unless @$first == @$second; 
   	for (my $i = 0; $i < @$first; $i++) { 
    	return 0 if $first->[$i] ne $second->[$i]; 
    } 
    return 1; 
}



#prints all element of the input array
#INPUT: string and 2d-array
sub print_2d_array{
	my $output_string=$_[0];
	my @array_2d=@{$_[1]};
	for my $elem1 ( 0 .. $#array_2d ) { 
		for my $elem2 ( 0 .. $#{$array_2d[$elem1]} ) {
			print "$output_string $elem1 -> $elem2 : $array_2d[$elem1]->[$elem2]\n";
		}
	}
	print "\n";
	return;
}



#opens a file and reads in the string-notation of a rooted binary phylogenetic tree if the notation fits in only one line
#INPUT: filename
#OUTPUT: array
sub read_tree_string{
  my $filename=$_[0];
  open (TREE, "<$filename") or die "cannot open $filename\n";
  my $i=0;
  my @tree;
  while (<TREE>){
    if ($i==0){
      chomp;
      $i++;
      @tree=split(/\s+/, $_);
      foreach my $elem (@tree){
        $elem=remove_commas($elem);
      }
    }
    elsif (($i==1)&&($_ eq "")){
      $i++;
    }
    elsif ($i>1){
      die "tree notation must fit in a single line\n";
    }
  }
  close TREE;
  @tree=tree::sort_tree(\@tree);
  return @tree;
}



#transforms an array into a hash
#INPUT: array
#OUTPUT: hash
sub array2hash {
    my @array=@{$_[0]};
    my %hash;
    my $i=1;
    foreach (@array) {
        $hash{$_}=$i++;
    }
    return %hash;
}



#checks if every array elem exists in hash
#INPUT: an array and a hash
#OUTPUT: bool
sub is_array_in_hash{
	my @array=@{$_[0]};
	my %hash=%{$_[1]};
	my $is_array=1;
	my $i=0;
	while (($is_array==1)&&($i<@array)){
		if (!exists $hash{$array[$i]}){
			$is_array=0;
		}
		else{
			$i++;
	  	}	
	}
	return $is_array;	
}



#deletes leaf-taxa in every node of tree including leaf itself
#e.g. tree: 1,2,3,4, 1,2 3,4 1 2    leaf: 3,4
#outputtree: 1,2 1 2
#if you call this function with a tree and any node, no matter if it is a leaf or not,
#all taxa of the node are deleted in all tree elements
#e.g tree: 1,2,3,4 1,2 3,4 1 2  node 1,2
#outputtree: 3,4
#INPUT: tree-array, leaf
#OUTPUT: tree-array
sub delete_part_of_node{
	my $leaf=$_[0];
	my @tree=@{$_[1]};
	my @taxa=split(/\,/, $leaf);
	foreach my $elem (@tree){
      	foreach my $elem2 (@taxa){
	      	$elem=substitute_node($elem, $elem2, "");
  		}
	}
	@tree=delete_empty_elements(\@tree);   #includes tree sort
	@tree=delete_double_elements(\@tree);
	return (@tree);
}



#deletes a node in a tree
#INPUT: node, tree-array
#OUTPUT: tree-array
sub delete_node{
	my $node=$_[0];
	my @tree=@{$_[1]};
	my $i=0;
	my @new_tree;
	foreach my $elem (@tree){
		if ($elem ne $node){
			$new_tree[$i]=$elem;
			$i++;
		}
	}
	return @new_tree;
}



#checks if all characters (seperated by a comma) of 'small' are in 'big'
#there is no need that the characters are sorted
#INPUT: two character (first: big, second:small)
#OUTPUT: boolean
sub is_in{
	my $big=$_[0];
	my $small=$_[1];
        my @big_taxa=split(/\,/, $big);
	my %big_taxa;
	foreach $elem (@big_taxa){   #could be replaced with array2hash function
		$big_taxa{$elem}=0;
	}
	my $bool=1;
        my @character=split(/\,/, $small);
	foreach my $elem (@character){
		if (not exists $big_taxa{$elem}){
			$bool=0;
		}
	}
	return $bool;	
}



#checks if node is a leaf
#works also for the root "#"
#INPUT: tree-array and node
#OUTPUT: bool
sub is_leaf{
	my $node=$_[0];
	my @tree=@{$_[1]};
	my @sons=sons($node, \@tree);
	if ((!defined($sons[0]) || $sons[0] eq "")){
		return 1;
	}
	else{
		return 0;
	}
}



#inserts root
#e.g. input: 1,2,3,4 1,2 3,4 1 2 3 4
#     output: #,1,2,3,4 1,2,3,4 1,2 3,4 1 2 3 4
#INPUT: tree-array
#OUTPUT: tree-array
sub insert_root{
	my @tree=@{$_[0]};
	my @tree_rooted;
	$tree_rooted[0]="#,".$tree[0];
	for (my $i=0; $i<@tree; $i++){
		$tree_rooted[$i+1]=$tree[$i];
	}
	return @tree_rooted;
}



#prints every element of the input array
#INPUT: array and string
sub print_array{
	my $output_string=$_[0];
	my @array=@{$_[1]};
	foreach my $elem (@array){
		print "$output_string $elem\n";
	}
	print "\n";
	return;
}



#deletes all characters of the short string in the long string, if present
#INPUT: two strings
#OUTPUT: string
sub delete_part_of_string{
	my $long_string=$_[0];
	my $short_string=$_[1];
	@to_delete=split(/\,/, $short_string);
	foreach my $elem (@to_delete){
		$long_string=substitute_node($long_string, $elem, "");
	}
	return $long_string;
}



#counts the numbers of leaves in a tree
#INPUT: tree-array
#OUTPUT: integer
sub count_leaves{
	my @tree=@{$_[0]};
	my $no=0;
	foreach my $node (@tree){
		if(is_leaf($node, \@tree)){
			$no++;
		}		
	}
	return $no;
}



#checks if graph is connected
#INPUT: tree-array
#OUTPUT: 1 if graph is connected, otherwise 0
sub is_connected{
	my @tree=@{$_[0]};
	my $graph=Graph::Undirected->new;
	
	for my $elem(@tree) { 
		$graph->add_vertex($elem);
		my @son=sons($elem, \@tree);
		if ($son[0] ne ""){
			$graph->add_edge($elem, $son[0]);
			if ($son[1] ne ""){
				$graph->add_edge($elem, $son[1]);
			}
		}
	}
	return $graph->is_connected;
}



#sorts a single node of the tree
#if there is a root character in the node it will be automatically sorted to the first position
#INPUT: a tree-node (string)
#OUTPUT: node
sub sort_node{  #numeric sort
	my $node=$_[0];
	my $bool=0;
	
	if ($node=~/#/){
		if ($node eq "#"){
		return $node;
		}
		else{
			$bool=1;
			$node=~s/#\,//;
		}
	}
	my @taxa=split(/\,/, $node);
	@taxa= sort { $a <=> $b } @taxa;
	$node = join(",",@taxa);
        if ($bool==1){
	    $node="#,".$node;
        }
	return $node;
}



#sorts the whole tree-array (1. legth, 2. alphanumerically
#INPUT: tree-array
#OUTPUT: tree-array
sub sort_tree{
	my @tree=@{$_[0]};
	foreach my $node (@tree){
		$node=sort_node($node);
	}
	@tree=sort(lengthsort @tree);
	return @tree;
	
}



#this function 'belongs' to the sort_tree function, sorting the nodes per length
sub lengthsort{
	my @no_a;
	my @no_b;
	if ($a!~/\,/){
		$no_a[0]=$a;	
	}
	else{
		@no_a=split(/\,/, $a);
	}
	if ($b!~/\,/){
		$no_b[0]=$b;	
	}
	else{
		@no_b=split(/\,/, $b);
	}
	if (@no_a==@no_b){
		my $i=0;
		while($i<@no_a){
			$no_a[$i]=~s/#/0/;
			$no_b[$i]=~s/#/0/;
			if($no_a[$i]>$no_b[$i]){
				#swap
				return 1;
			}
			elsif($no_a[$i]==$no_b[$i]){
				$i++;
			}
			else{
				return -1;
			}
		}
		return 0;
	}
	
	
	elsif (@no_a > @no_b){
		return -1;  #-1
	}
	else{
		#swap
		return 1; #1
	}
}



#remove double commas or a comma at the end or beginning of the string
#e.g. input ,1,2,,,4,5,
#     output 1,2,4,5
#INPUT: string
#OUTPUT: string
sub remove_commas{
	my $string=$_[0];
	$string=~s/\,+/\,/g;
	$string=~s/\,$//;
	$string=~s/^\,//;
	return $string;
}



#counts the cherries of a tree
#doesn't work for trees with node "#" for root!!!!!!!!!!!!
#INPUT: first node in tree-array ("0" as the starting number for cherries) and a tree-array
#OUTPUT: integer
sub count_cherries{
	my $node=$_[0];
	my $no=$_[1];
	my @tree=@{$_[2]};
	my @sons=sons($node, \@tree);
	if(is_leaf($node, \@tree)){
		return $no;
	}
	elsif((is_leaf($sons[0], \@tree))&&(is_leaf($sons[1], \@tree))){
		$no++;
		return $no
	}
	else{
		return(count_cherries($sons[0], $no, \@tree)+count_cherries($sons[1], $no, \@tree));
	}
}



#gives back the parent node of an input node in the tree-array
#parent of extra root(#) not considered so far
#INPUT: tree-array and node
#OUTPUT: node
sub get_parent{
  my $node=$_[0];
  my @tree=@{$_[1]};
  my $bool=0;
  my $parent=0;
  @tree=sort_tree(\@tree);
  $node=sort_node($node);
  my $i=@tree-1;
  while (($bool<=1)&&($i>=0)){
    if (($bool==1)&&(is_in($tree[$i], $node))){
      $parent=$tree[$i];
      $bool++;
    }
    if (($bool==0)&&($tree[$i] eq $node)){
      $bool++;
    }
    $i--;
  }
  if ($parent ne "0"){
    return $parent;
  }
  else{
    print "tree::get_parent ERROR no parent for $node found\n";
    die;
  }
}



#gets all leaves in a tree
#INPUT: tree-array
#OUTPUT: array (containing all leaves)
sub get_leaves{
  my @tree1=@{$_[0]};
  my @leaf;
  foreach my $elem (@tree1){
    if(is_leaf($elem, \@tree1)){  
          push(@leaf, $elem);
    }
  }
  return @leaf;
}


#calculates height of tree (longest path???????)
#height = length of the path (number of edges) from the root node to its furthest leaf
sub get_height{
  my @tree=@{$_[0]};
  my $max_height=0;
  for (my $i=0; $i<@tree; $i++){
    if(is_leaf($tree[$i], \@tree)){
      my $curr_height=0;
      for (my $j=0; $j<$i; $j++){
        if (is_in($tree[$j], $tree[$i])){
	  $curr_height++;
	}
      }
      if ($curr_height>$max_height){
        $max_height=$curr_height;
      }
    }
  }
  return $max_height
}


#calculates the depth of a single vertex of the tree
#depth = length of the path (number of edges) from the root to a single node
sub get_depth{
  my $node=$_[0];
  my @tree=@{$_[1]};
  my $i=0;
  my $depth=0;
  while (($i<@tree)&&($tree[$i] ne $node)){
    if (is_in($tree[$i], $node)){
      $depth++;
    }
    $i++;
  }
  
  return $depth;
}

1;
